@@ -10,7 +10,6 @@ import re
 import tarfile
 import time
 
-from io import BytesIO
 from json import loads
 from pathlib import Path
 from tempfile import TemporaryDirectory
@@ -26,33 +25,23 @@ from docker_registry_client_async import (
     ImageName,
 )
 from docker_sign_verify import ImageConfig, RegistryV2Manifest, RegistryV2ImageSource
-from docker_sign_verify.utils import be_kind_rewind, chunk_file
+from docker_sign_verify.utils import be_kind_rewind
 from pretty_bad_protocol import GPG
 from pretty_bad_protocol._parsers import Verify
 from pretty_bad_protocol._util import _make_binary_stream
 
 from .atomicsignature import AtomicSignature
+from .exceptions import DigestMismatchError, NoSignatureError, SignatureMismatchError
 from .imagestream import ImageStream
 from .singleassignment import SingleAssignment
 from .specs import OpenShiftReleaseSpecs
+from .utils import read_from_tar
 
 pytestmark = [pytest.mark.asyncio]
 
 LOGGER = logging.getLogger(__name__)
 
 
-async def read_from_tar(tar_file, tarinfo: tarfile.TarInfo) -> bytes:
-    """Reads an entry from a tar file into memory."""
-    bytesio = BytesIO()
-    await chunk_file(
-        tar_file.extractfile(tarinfo),
-        bytesio,
-        file_in_is_async=False,
-        file_out_is_async=False,
-    )
-    return bytesio.getvalue()
-
-
 class TypingDetachedSignature(NamedTuple):
     # pylint: disable=missing-class-docstring
     atomic_signature: AtomicSignature
@@ -102,24 +91,31 @@ class TypingSearchLayer(NamedTuple):
     release_metadata: Any
 
 
+class TypingVerifyDetachedSignature(NamedTuple):
+    # pylint: disable=missing-class-docstring
+    atomic_signature: AtomicSignature
+    crypt: Verify
+    result: bool
+
+
 class TypingVerifyReleaseMetadata(NamedTuple):
     # pylint: disable=missing-class-docstring
-    result: bool
     signatures: List[TypingDetachedSignature]
 
 
 async def _collect_digests(
+    *,
+    image_references: ImageStream,
     registry_v2_image_source: RegistryV2ImageSource,
     release_image_name: ImageName,
-    image_references: ImageStream,
 ) -> TypingCollectDigests:
     """
     Retrieves all blob and manifest digests for a given release.
 
     Args:
+        image_references: The image references for the release.
         registry_v2_image_source: The underlying registry v2 image source to use to retrieve the digests.
         release_image_name: The name of the release image.
-        image_references: The image references for the release.
 
     Returns:
         blobs: The mapping of blob digests to image prefixes.
@@ -134,7 +130,9 @@ async def _collect_digests(
         blobs[_digest].add(i_prefix)
 
     # TODO: Should we split out manifest and blob processing to separate functions?
-    for image_name, name in _get_tag_mapping(release_image_name, image_references):
+    for image_name, name in _get_tag_mapping(
+        release_image_name=release_image_name, image_references=image_references
+    ):
         # Convert tags to digests
         # pkg/cli/image/mirror/mirror.go:437 - plan()
         digest = image_name.digest
@@ -163,10 +161,11 @@ async def _collect_digests(
 
 
 async def _copy_blob(
-    registry_v2_image_source: RegistryV2ImageSource,
-    image_name_src: ImageName,
-    image_name_dest: ImageName,
+    *,
     digest: FormattedSHA256,
+    image_name_dest: ImageName,
+    image_name_src: ImageName,
+    registry_v2_image_source: RegistryV2ImageSource,
 ):
     LOGGER.debug("Copying blob %s ...", digest)
     if await registry_v2_image_source.layer_exists(image_name_dest, digest):
@@ -185,9 +184,10 @@ async def _copy_blob(
 
 
 async def _copy_manifest(
-    registry_v2_image_source: RegistryV2ImageSource,
-    image_name_src: ImageName,
+    *,
     image_name_dest: ImageName,
+    image_name_src: ImageName,
+    registry_v2_image_source: RegistryV2ImageSource,
 ):
     LOGGER.debug("Copying manifest to: %s ...", image_name_dest)
     LOGGER.debug("    from : %s", image_name_src)
@@ -221,7 +221,7 @@ async def _copy_manifest(
 
 
 async def _get_image_references(
-    tar_file, tarinfo: tarfile.TarInfo, path: Path
+    *, tar_file, tarinfo: tarfile.TarInfo, path: Path
 ) -> Optional[ImageStream]:
     """
     Retrieves images references from a given tarinfo, if available.
@@ -242,7 +242,7 @@ async def _get_image_references(
 
 
 async def _get_release_metadata(
-    tar_file, tarinfo: tarfile.TarInfo, path: Path
+    *, tar_file, tarinfo: tarfile.TarInfo, path: Path
 ) -> Optional[Any]:
     """
     Retrieves release metadata from a given tarinfo, if available.
@@ -262,7 +262,7 @@ async def _get_release_metadata(
         return loads(result)
 
 
-async def _get_request_headers(headers: LooseHeaders = None) -> LooseHeaders:
+async def _get_request_headers(*, headers: LooseHeaders = None) -> LooseHeaders:
     """
     Generates request headers that contain the user agent identifier.
 
@@ -285,7 +285,7 @@ async def _get_request_headers(headers: LooseHeaders = None) -> LooseHeaders:
 
 
 async def _get_security_information(
-    tar_file, tarinfo: tarfile.TarInfo, path: Path
+    *, tar_file, tarinfo: tarfile.TarInfo, path: Path
 ) -> TypingGetSecurityInformation:
     """
     Retrieves security information from a given tarinfo, if available.
@@ -336,7 +336,7 @@ async def _get_security_information(
 
 
 def _get_tag_mapping(
-    release_image_name: ImageName, image_references: ImageStream
+    *, release_image_name: ImageName, image_references: ImageStream
 ) -> Tuple[ImageName, str]:
     """
     Deconstructs the metadata inside an ImageStream into a mapping of image names to tag names.
@@ -357,7 +357,7 @@ def _get_tag_mapping(
         yield image_name, name
 
 
-def _import_owner_trust(gpg: GPG, trust_data: str):
+def _import_owner_trust(*, gpg: GPG, trust_data: str):
     # pylint: disable=protected-access
     """
     Imports trust information.
@@ -374,17 +374,18 @@ def _import_owner_trust(gpg: GPG, trust_data: str):
 
 
 async def _search_layer(
+    *,
+    layer: FormattedSHA256,
     registry_v2_image_source: RegistryV2ImageSource,
     release_image_name: ImageName,
-    layer: FormattedSHA256,
 ) -> TypingSearchLayer:
     """
     Searches image layers in a given release image for metadata.
 
     Args:
+        layer: The image layer to be searched.
         registry_v2_image_source: The underlying registry v2 image source to use to retrieve the metadata.
         release_image_name: The name of the release image.
-        layer: The image layer to be searched.
 
     Returns:
         image_references: An ImageStream object containing the image references, or None.
@@ -417,13 +418,19 @@ async def _search_layer(
                         # pkg/cli/image/extract/extract.go:616 - changeTarEntryParent()
                         # LOGGER.debug("Exclude %s due to missing prefix %s", path)
                         continue
-                    tmp = await _get_image_references(tar_file_in, tarinfo, path)
+                    tmp = await _get_image_references(
+                        tar_file=tar_file_in, tarinfo=tarinfo, path=path
+                    )
                     if tmp:
                         image_references.set(tmp)
-                    tmp = await _get_release_metadata(tar_file_in, tarinfo, path)
+                    tmp = await _get_release_metadata(
+                        tar_file=tar_file_in, tarinfo=tarinfo, path=path
+                    )
                     if tmp:
                         release_metadata.set(tmp)
-                    tmp = await _get_security_information(tar_file_in, tarinfo, path)
+                    tmp = await _get_security_information(
+                        tar_file=tar_file_in, tarinfo=tarinfo, path=path
+                    )
                     keys.extend(tmp.keys)
                     locations.extend(tmp.locations)
     return TypingSearchLayer(
@@ -434,11 +441,68 @@ async def _search_layer(
     )
 
 
+async def _verify_detached_signature(
+    *, digest: FormattedSHA256, gpg: GPG, signature: bytes
+) -> TypingVerifyDetachedSignature:
+    """
+    Verifies a detached signature against a given digest.
+
+    Args:
+        digest: Digest of the registry manifest.
+        gpg: GnuPG object to use to decrypt the signature.
+        signature: The signature to be verified
+
+    Returns:
+        The associated metadata used for verification.
+    """
+
+    # pkg/verify/verify.go:305: - verifySignatureWithKeyring()
+    LOGGER.debug("Verifying signature against digest: %s", digest)
+    crypt = gpg.decrypt(signature)
+    if not crypt.valid or crypt.trust_level < Verify.TRUST_ULTIMATE:
+        raise SignatureMismatchError("Signature does not match!")
+
+    LOGGER.debug("Signature matches:")
+    LOGGER.debug("  key       : %s", crypt.pubkey_fingerprint)
+    timestamp = time.strftime(
+        "%Y-%m-%d %H:%M:%S", time.gmtime(float(crypt.sig_timestamp))
+    )
+    LOGGER.debug("  timestamp : %s", timestamp)
+    LOGGER.debug("  username  : %s", crypt.username)
+
+    # pkg/verify/verify.go:382: - verifyAtomicContainerSignature()
+    atomic_signature = AtomicSignature(crypt.data)
+    if atomic_signature.get_type() != AtomicSignature.TYPE:
+        raise SignatureMismatchError("Signature is not the correct type!")
+
+    if len(atomic_signature.get_docker_reference().image) < 1:
+        raise SignatureMismatchError("Signature must have an identity!")
+
+    if atomic_signature.get_docker_manifest_digest() != digest:
+        raise DigestMismatchError("Signature digest does not match!")
+
+    LOGGER.debug("Signature is compliant:")
+    LOGGER.debug("  type                   : %s", atomic_signature.get_type())
+    LOGGER.debug(
+        "  docker reference       : %s",
+        atomic_signature.get_docker_reference(),
+    )
+    LOGGER.debug(
+        "  docker manifest digest : %s",
+        atomic_signature.get_docker_manifest_digest(),
+    )
+
+    return TypingVerifyDetachedSignature(
+        atomic_signature=atomic_signature, crypt=crypt, result=True
+    )
+
+
 async def _verify_release_metadata(
-    registry_v2_image_source: RegistryV2ImageSource,
+    *,
     digest: FormattedSHA256,
-    locations: List[str],
     keys: List[str],
+    locations: List[str],
+    registry_v2_image_source: RegistryV2ImageSource,
 ) -> TypingVerifyReleaseMetadata:
     # pylint: disable=protected-access,too-many-locals
     """
@@ -481,7 +545,7 @@ async def _verify_release_metadata(
         for key in gpg.list_keys():
             # TODO: Is it worth it to define pretty_bad_protocol._parsers.ImpoortOwnerTrust to validate
             #       the return value? ... or should we rely solely on crypt.trust_level below?
-            _import_owner_trust(gpg, f"{key['fingerprint']}:6:\n")
+            _import_owner_trust(gpg=gpg, trust_data=f"{key['fingerprint']}:6:\n")
 
         for key in gpg.list_keys():
             LOGGER.debug(
@@ -510,64 +574,31 @@ async def _verify_release_metadata(
                 LOGGER.debug("Signature retrieved.")
                 signature = await response.read()
 
-                # pkg/verify/verify.go:305: - verifySignatureWithKeyring()
-                LOGGER.debug("Verifying signature against digest: %s", digest)
-                crypt = gpg.decrypt(signature)
-                if not crypt.valid or crypt.trust_level < Verify.TRUST_ULTIMATE:
-                    LOGGER.error("Signature does not match!")
-                    continue
-
-                LOGGER.debug("Signature matches:")
-                LOGGER.debug("  key       : %s", crypt.pubkey_fingerprint)
-                timestamp = time.strftime(
-                    "%Y-%m-%d %H:%M:%S", time.gmtime(float(crypt.sig_timestamp))
+                metadata = await _verify_detached_signature(
+                    digest=digest, gpg=gpg, signature=signature
                 )
-                LOGGER.debug("  timestamp : %s", timestamp)
-                LOGGER.debug("  username  : %s", crypt.username)
-
-                # pkg/verify/verify.go:382: - verifyAtomicContainerSignature()
-                atomic_signature = AtomicSignature(crypt.data)
-                if atomic_signature.get_type() != AtomicSignature.TYPE:
-                    LOGGER.error("Signature is not the correct type!")
+                if not metadata.result:
                     continue
 
-                if len(atomic_signature.get_docker_reference().image) < 1:
-                    LOGGER.error("Signature must have an identity!")
-                    continue
-
-                if atomic_signature.get_docker_manifest_digest() != digest:
-                    LOGGER.error("Signature digest does not match!")
-                    continue
-
-                LOGGER.debug("Signature is compliant:")
-                LOGGER.debug(
-                    "  type                   : %s", atomic_signature.get_type()
-                )
-                LOGGER.debug(
-                    "  docker reference       : %s",
-                    atomic_signature.get_docker_reference(),
-                )
-                LOGGER.debug(
-                    "  docker manifest digest : %s",
-                    atomic_signature.get_docker_manifest_digest(),
-                )
-
                 result = True
                 signatures.append(
                     TypingDetachedSignature(
-                        atomic_signature=atomic_signature,
-                        key=crypt.pubkey_fingerprint,
+                        atomic_signature=metadata.atomic_signature,
+                        key=metadata.crypt.pubkey_fingerprint,
                         raw_signature=signature,
                         timestamp=time.strftime(
-                            "%Y-%m-%d %H:%M:%S", time.gmtime(float(crypt.sig_timestamp))
+                            "%Y-%m-%d %H:%M:%S",
+                            time.gmtime(float(metadata.crypt.sig_timestamp)),
                         ),
-                        verify=crypt,
+                        verify=metadata.crypt,
                         url=url,
-                        username=crypt.username,
+                        username=metadata.crypt.username,
                     )
                 )
 
-    return TypingVerifyReleaseMetadata(result=result, signatures=signatures)
+    if not result:
+        raise NoSignatureError("Unable to locate a valid signature!")
+    return TypingVerifyReleaseMetadata(signatures=signatures)
 
 
 async def get_release_metadata(
@@ -626,7 +657,9 @@ async def get_release_metadata(
     release_metadata = SingleAssignment("release_metadata")
     for layer in manifest.get_layers(None):
         tmp = await _search_layer(
-            registry_v2_image_source, release_image_name, FormattedSHA256.parse(layer)
+            layer=FormattedSHA256.parse(layer),
+            registry_v2_image_source=registry_v2_image_source,
+            release_image_name=release_image_name,
         )
         if tmp.image_references:
             image_references.set(tmp.image_references)
@@ -652,10 +685,11 @@ async def get_release_metadata(
         if signature_stores:
             _locations = signature_stores
         response = await _verify_release_metadata(
-            registry_v2_image_source, manifest_digest, _locations, _keys
+            digest=manifest_digest,
+            keys=_keys,
+            locations=_locations,
+            registry_v2_image_source=registry_v2_image_source,
         )
-        if not response.result:
-            raise RuntimeError("Release verification failed!")
         signatures = response.signatures
     else:
         LOGGER.debug("Skipping source release authenticity verification!")
@@ -664,7 +698,9 @@ async def get_release_metadata(
     assert image_references.get_json().get("apiVersion", "") == "image.openshift.io/v1"
 
     tmp = await _collect_digests(
-        registry_v2_image_source, release_image_name, image_references
+        image_references=image_references,
+        registry_v2_image_source=registry_v2_image_source,
+        release_image_name=release_image_name,
     )
     LOGGER.debug(
         "Collected %d manifests with %d blobs.",
@@ -779,14 +815,12 @@ async def put_release(
     )
 
     if verify:
-        response = await _verify_release_metadata(
-            registry_v2_image_source,
-            release_metadata.manifest_digest,
-            release_metadata.signature_stores,
-            release_metadata.signing_keys,
+        await _verify_release_metadata(
+            digest=release_metadata.manifest_digest,
+            keys=release_metadata.signing_keys,
+            locations=release_metadata.signature_stores,
+            registry_v2_image_source=registry_v2_image_source,
         )
-        if not response.result:
-            raise RuntimeError("Release verification failed!")
     else:
         LOGGER.debug("Skipping source release authenticity verification!")
 
@@ -809,7 +843,10 @@ async def put_release(
                 LOGGER.debug("    from : %s", image_name_src)
                 LOGGER.debug("    to   : %s", image_name_dest)
             await _copy_blob(
-                registry_v2_image_source, image_name_src, image_name_dest, digest
+                digest=digest,
+                image_name_dest=image_name_dest,
+                image_name_src=image_name_src,
+                registry_v2_image_source=registry_v2_image_source,
             )
             last_image_name_dest = image_name_dest
             last_image_name_src = image_name_src
@@ -822,7 +859,11 @@ async def put_release(
             endpoint=mirror_image_name.endpoint,
             tag=release_metadata.manifests[image_name_src],
         )
-        await _copy_manifest(registry_v2_image_source, image_name_src, image_name_dest)
+        await _copy_manifest(
+            image_name_dest=image_name_dest,
+            image_name_src=image_name_src,
+            registry_v2_image_source=registry_v2_image_source,
+        )
 
 
 async def translate_release_metadata(
